"""
Works roughly as follows:
1. Overall, we script a POST request (RQmain) to a single, complex form to create a Sakai Assignment.
2. However, the form includes one or more uploaded documents and those uploads (must) happen upfront
   (RQupload).
3. Worse, the overall form must be sent upfront of even the uploads (RQmain.0)!
   At least that is what happens when one opens the upload dialog. A symptom is the fact that one
   cannot open the upload dialog unless one has entered a title for the Assignment.
4. This appears to be needed because the URL created for each uploaded document contains an ID
   for the assignment.
5. Reverse engineering the details of this is difficult, because GETting the page on which the form resides
   (RQmain.get) involves about 40 requests.
6. Therefore, we do not yet know where the identitity of the Assignment freshly created in step 3
   is to be found. Our best guess is: Somewhere in the resulting HTML page.
7. Likewise for the URLs of the uploaded documents: They have to be sent as part of the big
   (>80 parameters) RQmain request, but the URL is not a main outcome of RQupload.   
8. Although the upload dialog is a separate window on top of the RQmain form, RQupload does not
   only close that window, it also reloads the RQmain page, presumably to include the freshly uploaded
   file in the list of files shown thereon.

"""
import collections.abc
import datetime as dt
import mimetypes
import os.path
import random
import re
import sys
import typing as tg
import urllib.parse

import bs4
import yaml
import requests

import base as b


usage = """mycampusscript create_multigroup_assgmt config.yaml
  Creates an Sakai 'assignment' entry in a site with separate deadlines per section
"""
site_url_help = """site_url: URL of the MyCampus site's homepage"""
cookies_help = """cookies: a dictionary of cookie names to cookie values
  You need to obtain the following cookies: SAKAI2SESSIONID.  How to do that:
  Log into MyCampus, 
  visit the site where you want to create the multigroup assignment,
  open the developer tools in your browser (in Firefox: F12),
  open the 'network' tab in the developer tools,
  visit the assignment creation form ('Assignments' -> 'Add'),
  select the first requests in the network tab's request list,
  in the panel on the right, select 'Headers' -> 'Request Headers', turn on 'raw' view,
  copy the entire line that starts with "Cookie:" and paste into your config as the value
    for the 'cookies' config parameter; you must enclose it in double quotes.
"""
group_ids_help = """group_ids: a dictionary of group_id->groupname as indicated in the respective
  MyCampus form page HTML. The values the script has found are the following (you can copy-and-paste
  this directly into your config file):
group_ids:
{group_ids_yaml}
"""
assignment_title_help = """assignment_title: a string as would be provided in the form"""
assignment_instructions_help = """assignment_instructions: an HTML fragment (at least <p>some instructions text</p>)"""
attachments_help = """attachments: list of filenames (the basename will be visible in the assignment) like this:
attachments:
  - mypath/file1.pdf
  - otherpath/image.jpg
"""
open_date_help = """open_date: A date and time string when the assignment will become visible, e.g.
open_date: 2024-12-13 14:44
"""
base_due_date_help = """base_due_date: yyyy-mm-dd  (due dates will be computed relative to midnight of that day)"""
accept_until_additional_minutes_help = """accept_until_additional_minutes: number of minutes after due_date
that the assignment stays open, e.g.
accept_until_additional_minutes: 5
"""
relative_due_dates_help = """relative_deadlines: mapping from group_id to relative time. You can copy-paste-modify
the following with true group_ids and random times:
relative_deadlines:
{relative_deadlines_yaml}
"""
send_a_reminder_help = """send_a_reminder: flag, see the web dialog, e.g.
send_a_reminder: false
"""
hide_due_date_from_students_help = """hide_due_date_from_students: flag, see the web dialog"""
add_due_date_to_calendar_help = """add_due_date_to_calendar: flag, see the web dialog"""
add_an_announcement_help = """add_an_announcement: flag, see the web dialog"""
submission_notification_email_help = """submission_notification_email: flag whether to send daily notifications,
see the web dialog, e.g.
submission_notification_email: true
"""
grade_this_assignment_help = """grade_this_assignment: flag.
If true, the grade scale type is always 'points'.
"""
max_points_help = """max_points: int, see the web dialog."""
gradebook_categories_help = """gradebook_categories: a dictionary of category_id->category_name as indicated 
  in the respective MyCampus form page HTML. The values the script has found are the following 
  (you can copy-and-paste this directly into your config file):
gradebook_categories:
{gradebook_categories_yaml}
  0: do not send anything to gradebook at all
"""
send_grades_to_gradebook_category_help = """send_grades_to_gradebook_category: code from the above table, e.g.
send_grades_to_gradebook_category: -1
"""
released_grade_notification_email_help = """released_grade_notification_email: bool, e.g.
released_grade_notification_email: false
"""


ASSIGNMENTS_CREATION_FORM_PATH = 'tool/1f8f054a-4e58-4965-8d32-141b236e0023?sakai_action=doNew_assignment'
ASSIGNMENTS_CREATION_ATTACHMENT_UPLOAD_PATH = 'tool/1f8f054a-4e58-4965-8d32-141b236e0023/sakai.filepicker.helper'
SESSIONCOOKIE_NAME = 'SAKAI2SESSIONID'
ASSIGNMENT_FORM_ID = 'newAssignmentForm'
CSRF_TOKEN_CONFIGENTRY_NAME = 'csrf_token'
CSRF_TOKEN_HTML_NAME = 'sakai_csrf_token'
RBCS_TOKEN_CONFIGENTRY_NAME = 'rbcs_token'
RBCS_TOKEN_HTML_NAME = 'rbcs-token'

StrStrDict = dict[str, str]
StrAnyDict = dict[str, tg.Any]
ReplacementsFunc = collections.abc.Callable[[StrAnyDict], StrStrDict]


def main(scriptname: str, cmdname: str, configfile: str):
    investigate_jsonlog("/c/temp/irgendwas3.json")
    return


    with open(configfile, 'rt') as f:
        config = yaml.load(f, Loader=yaml.Loader)
    site_url: str = find_value_or_help(config, 'site_url')
    parse_cookies(config)
    obtain_tokens(config)
    group_ids: StrStrDict = find_value_or_help(config, 'group_ids', dynamic=get_group_ids)
    assignment_title: str = find_value_or_help(config, 'assignment_title')
    assignment_instructions: str = find_value_or_help(config, 'assignment_instructions')
    attachments: list[str] = find_value_or_help(config, 'attachments')
    open_date: str = find_value_or_help(config, 'open_date')
    open_date_dt: dt.datetime = dt.datetime.strptime(open_date, "%Y-%m-%d %H:%M")
    base_due_date: dt.date = find_value_or_help(config, 'base_due_date')
    accept_until_additional_minutes: int = find_value_or_help(config, 'accept_until_additional_minutes')
    accept_until_td: dt.timedelta = dt.timedelta(minutes=accept_until_additional_minutes)
    relative_due_dates: StrStrDict = find_value_or_help(config, 'relative_due_dates', dynamic=get_deadlines)
    absolute_due_dates = compute_due_dates(base_due_date, relative_due_dates)
    # for dl in absolute_due_dates.values():
    #     print(dt.datetime.strftime(dl, "%Y-%m-%d %H:%M"))
    send_a_reminder: bool = find_value_or_help(config, 'send_a_reminder')
    hide_due_date_from_students: bool = find_value_or_help(config, 'hide_due_date_from_students')
    add_due_date_to_calendar: bool = find_value_or_help(config, 'add_due_date_to_calendar')
    add_an_announcement: bool = find_value_or_help(config, 'add_an_announcement')
    # TODO 3: ask for submission type
    # TODO 3: handle resubmission logic (complicated!)
    submission_notification_email: bool = find_value_or_help(config, 'submission_notification_email')
    grade_this_assignment: bool = find_value_or_help(config, 'grade_this_assignment')
    max_points: int = find_value_or_help(config, 'max_points')
    gradebook_categories: int = find_value_or_help(config, 'gradebook_categories',
                                                   dynamic=get_gradebook_categories)
    send_grades_to_gradebook_category: int = find_value_or_help(config, 'send_grades_to_gradebook_category')
    released_grade_notification_email: bool = find_value_or_help(config, 'released_grade_notification_email')
    for attachment in attachments:
        upload_attachment(config, attachment)


def find_value_or_help(config: StrAnyDict, key: str, dynamic: ReplacementsFunc = None) -> tg.Any:
    if key in config:
        return config[key]
    the_help = globals()[f"{key}_help"]
    if dynamic:
        replacements = dynamic(config)
        the_help = the_help.format(**replacements)
    b.critical(the_help)


def compute_due_dates(base: dt.date, relative: StrStrDict) -> dict[str, dt.datetime]:
    base_dt = dt.datetime(base.year, base.month, base.day)
    result = dict()
    for group_id, relative_dl in relative.items():
        mm = re.fullmatch(r"(\d+)d(\d+)h(\d+)m", relative_dl)
        if not mm:
            b.critical(f"Ill-formed relative deadline '{relative_dl}', should be like '4d14h0m'")
        days, hours, minutes = int(mm.group(1)), int(mm.group(2)), int(mm.group(3)) 
        delta = dt.timedelta(days=days, hours=hours, minutes=minutes)
        result[group_id] = base_dt + delta
    return result


def get_gradebook_categories(config: StrAnyDict) -> StrStrDict:
    soup = get_formpage_soup(config)
    categories = soup.find(id='category')
    lines = []
    for option in categories.find_all('option'):
        # <option value="706" >Notizzettel (insg. >= n-2)</option>
        _id = option['value']
        name = str(option.string)  # e.g. Notizzettel (insg. >= n-2)
        lines.append(f"  {_id}: \"{name}\"")
    return dict(gradebook_categories_yaml="\n".join(lines))


def get_group_ids(config: StrAnyDict) -> StrStrDict:
    """
    GET the form page and extract the list of group names and group IDs, called selectedGroups.
    Store the formpage as config['formpage_soup'].
    """
    soup = get_formpage_soup(config)
    options = soup.find(id='selectedGroups')
    lines = []
    for option in options.find_all('option'):
        # <option value="3fe41c9..."  >Groupname  (Coursename, Groupname) </option>
        _id = option['value']
        name = str(option.string)  # Groupname  (Coursename, Groupname)
        lines.append(f"  {_id}: \"{name}\"")
    return dict(group_ids_yaml="\n".join(lines))


def get_deadlines(config: StrAnyDict) -> StrStrDict:
    lines = []
    for group_id, group_name in config['group_ids']:
        deadline = "%dd%02dh00m" % (random.randint(1, 5), 2*random.randint(4, 8))
        lines.append(f"  {group_id}: {deadline}")
    return dict(relative_deadlines_yaml="\n".join(lines))


def get_formpage_soup(config: StrAnyDict) -> bs4.BeautifulSoup:
    """
    Retrieves content of form webpage as a soup from config if present or
    else GETs it first and stores it there.
    """
    if 'formpage_soup' in config:
        return config['formpage_soup']
    url = f"{config['site_url']}/{ASSIGNMENTS_CREATION_FORM_PATH}"
    sessioncookie = config['cookies'][SESSIONCOOKIE_NAME]
    response = requests.get(url, cookies={SESSIONCOOKIE_NAME: sessioncookie})
    assert response.status_code == 200
    config['formpage_soup'] = bs4.BeautifulSoup(response.text, 'html.parser')
    return config['formpage_soup']


def parse_cookies(config: StrAnyDict):
    """Sets config['cookies'] = dict(cookiename=value, ...)"""
    def perhaps_complain(mm):
        if not mm:
            b.critical("config param 'cookies' must have format 'Cookie: \"NAME1=value1; name2=value\"")

    cookiesheader_re = r"Cookie: (.+)"
    mm = re.fullmatch(cookiesheader_re, config['cookies'])
    perhaps_complain(mm)
    cookiepairs = mm.group(1).split("; ")
    result = dict()
    for pair in cookiepairs:
        mm = re.fullmatch(r"(.+?)=(.+)", pair)  # split at the first equals sign
        perhaps_complain(mm)
        result[mm.group(1)] = mm.group(2)
    # print(result)
    config['cookies'] = result


def obtain_tokens(config: StrAnyDict):
    """Find CSRF token and RBCS token in form web page and store them in config"""
    soup = get_formpage_soup(config)
    form = soup.find(id=ASSIGNMENT_FORM_ID)
    csrftoken_input_elem = form.find('input', attrs=dict(name=CSRF_TOKEN_HTML_NAME))
    rbcstoken_input_elem = form.find('input', attrs=dict(name=RBCS_TOKEN_HTML_NAME))
    config[CSRF_TOKEN_CONFIGENTRY_NAME] = csrftoken_input_elem['value']
    config[RBCS_TOKEN_CONFIGENTRY_NAME] = rbcstoken_input_elem['value']


def upload_attachment(config: StrAnyDict, filepath: str):
    """
    Make POST request with multipart/formdata body, many headers, and the following format of URL:
    POST /portal/site/<site-id>/tool/1f8f054a-4e58-4965-8d32-141b236e0023/sakai.filepicker.helper?special=upload&sakai_csrf_token=<hexstring>&sakai_action=doAttachupload
    See c:/temp/a.json log.entries[164].
    """
    base_url = f"{config['site_url']}/{ASSIGNMENTS_CREATION_ATTACHMENT_UPLOAD_PATH}"
    query_args = dict(special='upload', sakai_action='doAttachupload',  # panel='Main', ?
                      sakai_csrf_token=config[CSRF_TOKEN_CONFIGENTRY_NAME])
    # TODO 1: continue here:
    # We next need to construct the multipart body of the POST request.
    # See "POST Multiple Multipart-Encoded Files" in https://requests.readthedocs.io/en/latest/user/advanced/
    # We have so far failed to obtain the HTML of the form to be posted, which we need to find the
    # name of the attachment file form element.
    # in c:/temp/a.json:
    # Request 164 ist der POST-Request, den wir analog nachbilden müssen.
    # Das JSON ist dafür unpraktisch, weil der ganze multipart-Rumpf nur als Text dasteht.
    # In der Word-Datei findet sich die aufgeschlüsselte Form, die der Browser dafür anzeigt.
    # Erzeugung eines solchen Requests, siehe https://stackoverflow.com/a/12385661
    headers = dict(Referer=f"{config['site_url']}/{ASSIGNMENTS_CREATION_ATTACHMENT_UPLOAD_PATH}")
    filename = os.path.basename(filepath)
    content_type, encoding = mimetypes.guess_type(filename)
    multipart_body = dict(
        source='0',
        collectionId=f'/group/{site_id(config["site_url"])}',
        sakai_action='',
        navRoot='',
        criteria='title',
        rt_action='',
        selectedItemId='',
        upload=(filename, slurp(filepath), content_type),
        url='',
        sakai_csrf_token=config[CSRF_TOKEN_CONFIGENTRY_NAME]
    )
    multipart_body['from'] = 'list'  # separate because 'from' is a keyword
    resp = requests.post(base_url, 
                         params=query_args,
                         headers=headers,
                         files=multipart_body, 
                         cookies={SESSIONCOOKIE_NAME: config['cookies'][SESSIONCOOKIE_NAME]})
    print(resp.text)


def investigate_jsonlog(jsonfile: str):
    """
    Helper routine for reverse engineering:
    We record a complete RQmain-posting episode with two attachments in Firefox
    (F12->Network; cogwheel menu->Persist Logs; perform episode; SaveAs:irgendwas.json).
    We pull out parts needed for understanding here.
    """
    import json
    import dictns
    testdata = dictns.Namespace(dict(
        title="irgendwas3",
        a_id='dbb411b5-195d-4089-ae66-6e9c8f01b81c',
        instructions="irgendwelche Anleitung",
        attachments=["ajeunwuk", "akeunwuk"],
        group="Übung 01",  # name prefix only
    ))
    with open(jsonfile, 'r') as f:
        log = json.load(f)
    ns = dictns.Namespace(log)
    # the first RQmain POST request is #111, response is 302 to panel=Main
    firstrequest = ns.log.entries[111]
    assert firstrequest.request.postData.params[5].value == testdata.title
    # ns.log.entries[112] is another 302 to ASSIGNMENTS_CREATION_ATTACHMENT_UPLOAD_PATH
    # popuprequest = ns.log.entries[113]
    url = urllib.parse.urlparse("http://localhost")
    for i in range(len(ns.log.entries)):  # range(111, 178+1):
        entry = ns.log.entries[i]
        if "portal/site" not in entry.request.url:
            continue
        last_url = url
        url = urllib.parse.urlparse(entry.request.url)
        if url == last_url:
            msg = "(ditto)"
        elif url.path == last_url.path:
            msg = url.query
        else:
            msg = entry.request.url
        print(i, entry.response.status, msg)
            
    # print(popuprequest.response.content.text)


def site_id(site_url: str) -> str:
    """From e.g. https://qmycampus.imp.fu-berlin.de/portal/site/724a527a-db9d, retrieve 724a527a-db9d"""
    path = urllib.parse.urlparse(site_url).path
    the_id = os.path.basename(path)
    return the_id


def slurp(filename: str) -> bytes:
    with open(filename, 'rb') as f:
        return f.read()



if __name__ == '__main__':
    if len(sys.argv) != 2+1 or sys.argv[1] != 'create_multigroup_assgmt':
        print(usage)
    else:
        try:
            main(*sys.argv)
        except b.CritialError:
            pass  # just stop
